-- stylua: ignore start
import 'CoreLibs/animation.lua'
import 'CoreLibs/animator.lua'
import 'CoreLibs/easing.lua'
import 'CoreLibs/graphics.lua'
import 'CoreLibs/object.lua'
import 'CoreLibs/ui.lua'

import 'graphics.lua'
import 'utils.lua'
-- stylua: ignore end

--- @alias XYPair { x: number, y: number }

--- @alias Base { x: number, y: number }

--- @alias Runner { x: number, y: number, nextBase: Base, prevBase: Base, forcedTo: Base }

--- @alias Fielder { onArrive: fun() | nil, x: number | nil, y: number | nil, target: XYPair | nil, speed: number }

local gfx <const> = playdate.graphics

local Screen <const> = {
	W = playdate.display.getWidth(),
	H = playdate.display.getHeight(),
}

local Center <const> = xy(Screen.W / 2, Screen.H / 2)

local BatCrackSound <const> = playdate.sound.sampleplayer.new("sounds/bat-crack-reverb.wav")
local GrassBackground <const> = gfx.image.new("images/game/grass.png") --[[@as PlaydateGraphicsImage]]
local PlayerFrown <const> = gfx.image.new("images/game/player-frown.png") --[[@as PlaydateGraphicsImage]]

local PlayerImageBlipper <const> = blipper.new(100, "images/game/player.png", "images/game/player-lowhat.png")

local DanceBounceMs <const> = 500
local DanceBounceCount <const> = 4
local FielderDanceAnimator <const> = gfx.animator.new(DanceBounceMs, 10, 0, easingHill) --, -1 * DanceBounceMs * DanceBounceCount)
FielderDanceAnimator.repeatCount = DanceBounceCount - 1

local BallOffscreen <const> = 999

local PitchFlyMs <const> = 2500
local PitchStartX <const> = 200
local PitchStartY <const>, PitchEndY <const> = 90, 250

local PitchAnimator <const> = gfx.animator.new(PitchFlyMs, PitchStartY, PitchEndY, playdate.easingFunctions.outQuint)

local CrankOffsetDeg <const> = 90
local BatOffset <const> = xy(10, 25)

local batBase <const> = xy(Center.x - 34, 215)
local batTip <const> = xy(0, 0)

local TagDistance <const> = 20

local ball <const> = {
	x = Center.x,
	y = Center.y,
	size = 6,
}

local BatLength <const> = 45

local Modes <const> = {
	batting = {},
	running = {},
}

local currentMode = Modes.batting

local ballAnimatorY = gfx.animator.new(0, BallOffscreen, BallOffscreen, playdate.easingFunctions.linear)
local ballAnimatorX = gfx.animator.new(0, BallOffscreen, BallOffscreen, playdate.easingFunctions.linear)

-- TODO? Replace this AND ballSizeAnimator with a ballHeightAnimator
-- ...that might lose some of the magic of both. Compromise available? idk
local ballFloatAnimator <const> = gfx.animator.new(2000, -60, 0, easingHill)
local BallSizeMs = 2000
local ballSizeAnimator = gfx.animator.new(BallSizeMs, 9, 6, easingHill)

local HitMult = 10

local deltaSeconds = 0

local First <const>, Second <const>, Third <const>, Home <const> = 1, 2, 3, 4

---@type Base[]
local Bases = {
	xy(Screen.W * 0.93, Screen.H * 0.52),
	xy(Screen.W * 0.47, Screen.H * 0.19),
	xy(Screen.W * 0.03, Screen.H * 0.52),
	xy(Screen.W * 0.474, Screen.H * 0.79),
}

-- Pseudo-base for batter to target
local RightHandedBattersBox <const> = xy(Bases[Home].x - 35, Bases[Home].y)

---@type table<Base, Base>
local NextBaseMap <const> = {
	[Bases[First]] = Bases[Second],
	[Bases[Second]] = Bases[Third],
	[Bases[Third]] = Bases[Home],
}

function newFielder(speed)
	return {
		speed = speed,
	}
end

---@type table<string, Fielder>
local fielders <const> = {
	first = newFielder(40),
	second = newFielder(40),
	shortstop = newFielder(40),
	third = newFielder(40),
	pitcher = newFielder(30),
	catcher = newFielder(20),
	left = newFielder(40),
	center = newFielder(40),
	right = newFielder(40),
}

local PitcherStartPos <const> = {
	x = Screen.W * 0.48,
	y = Screen.H * 0.40,
}

--- Resets the target positions of all fielders to their defaults (at their field positions).
---@param fromOffTheField boolean If provided, also sets all runners' current position to one centralized location.
function resetFielderPositions(fromOffTheField)
	if fromOffTheField then
		for _, fielder in pairs(fielders) do
			fielder.x = Center.x
			fielder.y = Screen.H
		end
	end

	fielders.first.target = xy(Screen.W - 65, Screen.H * 0.48)
	fielders.second.target = xy(Screen.W * 0.70, Screen.H * 0.30)
	fielders.shortstop.target = xy(Screen.W * 0.30, Screen.H * 0.30)
	fielders.third.target = xy(Screen.W * 0.1, Screen.H * 0.48)
	fielders.pitcher.target = xy(PitcherStartPos.x, PitcherStartPos.y)
	fielders.catcher.target = xy(Screen.W * 0.475, Screen.H * 0.92)
	fielders.left.target = xy(Screen.W * -1, Screen.H * -0.2)
	fielders.center.target = xy(Center.x, Screen.H * -0.4)
	fielders.right.target = xy(Screen.W * 2, Screen.H * fielders.left.target.y)
end

local BatterStartingX <const> = Bases[Home].x - 40
local BatterStartingY <const> = Bases[Home].y - 3

--- @type Runner[]
local runners <const> = {}

--- @type Runner[]
local outRunners <const> = {}

---@return Runner
function newRunner()
	local new = {
		x = BatterStartingX - 60,
		y = BatterStartingY + 60,
		nextBase = RightHandedBattersBox,
		prevBase = nil,
		forcedTo = Bases[First],
	}
	runners[#runners + 1] = new
	return new
end

---@type Runner | nil
local batter = newRunner()

--- "Throws" the ball from its current position to the given destination.
function throwBall(destX, destY, easingFunc, flyTimeMs, floaty)
	if not flyTimeMs then
		flyTimeMs = distanceBetween(ball.x, ball.y, destX, destY) * 5
	end
	ballSizeAnimator:reset(flyTimeMs)
	ballAnimatorY = gfx.animator.new(flyTimeMs, ball.y, destY, easingFunc)
	ballAnimatorX = gfx.animator.new(flyTimeMs, ball.x, destX, easingFunc)
	if floaty then
		ballFloatAnimator:reset(flyTimeMs)
	end
end

function pitch()
	currentMode = Modes.batting
	ballAnimatorX = gfx.animator.new(0, PitchStartX, PitchStartX, playdate.easingFunctions.linear)
	ballAnimatorY = PitchAnimator
	PitchAnimator:reset()
end

function playdate.AButtonDown()
	if not batter then
		return
	end
	pitch()
end

function playdate.upButtonDown() end

function playdate.downButtonDown() end

function playdate.rightButtonDown() end

function playdate.leftButtonDown() end

local elapsedSec = 0
local crankChange = 0
local acceleratedChange

local BaseHitbox = 13
--- Returns the base being touched by the runner at (x,y), or nil, if no base is being touched
function isTouchingBase(x, y)
	for _, base in ipairs(Bases) do
		if distanceBetween(x, y, base.x, base.y) < BaseHitbox then
			return base
		end
	end

	return nil
end

local BallCatchHitbox = 3
function isTouchingBall(x, y)
	local ballDistance = distanceBetween(x, y, ball.x, ball.y)
	return ballDistance < BallCatchHitbox
end

local outs = 0
local homeScore = 0
local awayScore = 0

function outRunner(runnerIndex)
	outs = math.min(3, outs + 1)
	outRunners[#outRunners + 1] = runners[runnerIndex]
	table.remove(runners, runnerIndex)
	FielderDanceAnimator:reset()
end

-- TODO: Away score
function score(runnerIndex)
	outRunners[#outRunners + 1] = runners[runnerIndex]
	table.remove(runners, runnerIndex)
	homeScore = homeScore + 1
end

function updateFielders()
	local touchingBaseCache = buildCache(function(runner)
		return isTouchingBase(runner.x, runner.y)
	end)

	for _, fielder in pairs(fielders) do
		-- TODO: Target unforced runners (or their target bases) for tagging
		-- With new Position-based scheme, fielders are now able to set `fielder.target = runner` to track directly
		if fielder.target ~= nil then
			local x, y, distance = normalizeVector(fielder.x, fielder.y, fielder.target.x, fielder.target.y)

			if distance > 1 then
				fielder.x = fielder.x - (x * fielder.speed * deltaSeconds)
				fielder.y = fielder.y - (y * fielder.speed * deltaSeconds)
			else
				if fielder.onArrive then
					fielder.onArrive()
				end
				fielder.target = nil
			end
		end

		if currentMode == Modes.running and isTouchingBall(fielder.x, fielder.y) then
			local touchedBase = isTouchingBase(fielder.x, fielder.y)
			for i, runner in pairs(runners) do
				local runnerOnBase = touchingBaseCache.get(runner)
				if
					touchedBase
					and runner.prevBase
					and runner.forcedTo == touchedBase
					and touchedBase ~= runnerOnBase
				then
					outRunner(i)
				elseif not runnerOnBase then
					if distanceBetween(runner.x, runner.y, fielder.x, fielder.y) < TagDistance then
						outRunner(i)
					end
				end
			end
		end
	end
end

--- Returns true if at least one runner is still moving
---@return boolean
function updateRunners()
	local autoRunSpeed = 20
	--autoRunSpeed = 140
	local currentRunners = currentMode == Modes.batting and { batter }
		or filter(runners, function(runner)
			return runner ~= batter
		end)

	local runnerMoved = false
	for runnerIndex, runner in ipairs(currentRunners) do
		local nearestBase, nearestBaseDistance = getNearestOf(Bases, runner.x, runner.y)
		if
			nearestBaseDistance < 5
			and runner.prevBase
			and runner.nextBase == Bases[Home]
			and nearestBase == Bases[Home]
		then
			score(runnerIndex)
		end
		if runner.nextBase then
			local nb = runner.nextBase
			local x, y, distance = normalizeVector(runner.x, runner.y, nb.x, nb.y)

			if distance > 1 then
				local mult = 1
				if crankChange < 0 then
					mult = -1
				end
				local prevX, prevY = runner.x, runner.y
				-- TODO: Drift toward nearest base?
				local autoRun = nearestBaseDistance > 5 and mult * autoRunSpeed * deltaSeconds or 0
				mult = autoRun + (crankChange / 20)
				runner.x = runner.x - (x * mult)
				runner.y = runner.y - (y * mult)
				runnerMoved = runnerMoved or prevX ~= runner.x or prevY ~= runner.y
			else
				runner.nextBase = NextBaseMap[runner.nextBase]
				runner.forcedTo = nil
			end
		end
	end

	return runnerMoved
end

function getRunnerTargeting(base)
	for _, runner in pairs(runners) do
		if runner.nextBase == base then
			return runner
		end
	end
	return nil
end

---@return Base[]
function getForcedOutTargets()
	local targets = {}
	for _, base in ipairs(Bases) do
		local runnerTargetingBase = getRunnerTargeting(base)
		if runnerTargetingBase then
			targets[#targets + 1] = base
		else
			return targets
		end
	end
	return targets
end

--- Returns the position,distance of the basest closest to the runner furthest from a base
---@return { x: number, y: number } | nil, number | nil
function getBaseOfStrandedRunner()
	local farRunnersBase, farDistance
	for _, runner in pairs(runners) do
		local nearestBase, distance = getNearestOf(Bases, runner.x, runner.y, function(base)
			return runner.nextBase == base
		end)
		if farRunnersBase == nil or farDistance < distance then
			farRunnersBase = nearestBase
			farDistance = distance
		end
	end

	return farRunnersBase, farDistance
end

--- Returns x,y of the throw target
---@return number|nil, number|nil
function getNextThrowTarget()
	-- TODO: Handle missed throws, check for fielders at target, etc.
	local targets = getForcedOutTargets()
	if #targets ~= 0 then
		return targets[1].x, targets[1].y
	end

	local baseCloseToStrandedRunner = getBaseOfStrandedRunner()
	if baseCloseToStrandedRunner then
		return baseCloseToStrandedRunner.x, baseCloseToStrandedRunner.y
	end
end

local ResetFieldersAfterSeconds = 4
local secondsSinceLastRunnerMove = 0

local PitchAfterSeconds = 5
local secondsSincePitchAllowed = -5

function init()
	playdate.display.setRefreshRate(50)
	gfx.setBackgroundColor(gfx.kColorWhite)
	playdate.setMenuImage(gfx.image.new("images/game/menu-image.png"))
	resetFielderPositions(true)
	playdate.getSystemMenu():addMenuItem("Restart game", function() end)
end

function updateBatting()
	secondsSincePitchAllowed = secondsSincePitchAllowed + deltaSeconds
	if secondsSincePitchAllowed > PitchAfterSeconds then
		pitch()
		secondsSincePitchAllowed = 0
	end
	if ball.y < BallOffscreen then
		ball.y = ballAnimatorY:currentValue() + ballFloatAnimator:currentValue()
		ball.size = 6
	end

	local batAngle = math.rad(playdate.getCrankPosition() + CrankOffsetDeg)
	-- TODO: animate bat-flip or something
	batBase.x = batter and (batter.x + BatOffset.x) or 0
	batBase.y = batter and (batter.y + BatOffset.y) or 0
	batTip.x = batBase.x + (BatLength * math.sin(batAngle))
	batTip.y = batBase.y + (BatLength * math.cos(batAngle))

	if
		acceleratedChange >= 0
		and pointDirectlyUnderLine(ball.x, ball.y, batBase.x, batBase.y, batTip.x, batTip.y, Screen.H)
	then
		BatCrackSound:play()
		currentMode = Modes.running
		local ballAngle = batAngle + math.rad(90)

		local mult = math.abs(crankChange / 15)
		local ballVelX = mult * 10 * math.sin(ballAngle)
		local ballVelY = mult * 5 * math.cos(ballAngle)
		if ballVelY > 0 then
			ballVelX = ballVelX * -1
			ballVelY = ballVelY * -1
		end
		local ballDestX = ball.x + (ballVelX * HitMult)
		local ballDestY = ball.y + (ballVelY * HitMult)
		-- Hit!
		throwBall(ballDestX, ballDestY, playdate.easingFunctions.outQuint, 2000)

		fielders.first.target = Bases[First]
		batter.nextBase = Bases[First]
		batter.prevBase = Bases[Home]
		batter.forcedTo = Bases[First]
		batter = nil -- Demote batter to a mere runner

		local chasingFielder = getNearestOf(fielders, ballDestX, ballDestY)
		chasingFielder.target = { x = ballDestX, y = ballDestY }
		chasingFielder.onArrive = function()
			local targetX, targetY = getNextThrowTarget()
			if targetX ~= nil then
				throwBall(targetX, targetY, playdate.easingFunctions.linear, nil, true)
				chasingFielder.onArrive = nil
			end
		end
	end
end

function updateRunning()
	secondsSincePitchAllowed = 0
	ball.size = ballSizeAnimator:currentValue()
	if updateRunners() then
		secondsSinceLastRunnerMove = 0
	else
		secondsSinceLastRunnerMove = secondsSinceLastRunnerMove + deltaSeconds
		if secondsSinceLastRunnerMove > ResetFieldersAfterSeconds then
			throwBall(PitchStartX, PitchStartY, playdate.easingFunctions.linear, nil, true)
			resetFielderPositions(false)
			currentMode = Modes.batting
			batter = newRunner()
		end
	end
end

function updateOutRunners()
	for i, runner in ipairs(outRunners) do
		if runner.x < Screen.W + 50 and runner.y < Screen.H + 50 then
			runner.x = runner.x + (deltaSeconds * 25)
			runner.y = runner.y + (deltaSeconds * 25)
		else
			table.remove(outRunners, i)
		end
	end
end

function updateGameState()
	deltaSeconds = playdate.getElapsedTime() or 0
	playdate.resetElapsedTime()
	elapsedSec = elapsedSec + deltaSeconds
	crankChange, acceleratedChange = playdate.getCrankChange() --[[@as number, number]]

	ball.x = ballAnimatorX:currentValue()
	ball.y = ballAnimatorY:currentValue() + ballFloatAnimator:currentValue()

	if currentMode == Modes.batting then
		updateBatting()
		updateRunners()
	elseif currentMode == Modes.running then
		updateRunning()
	end

	updateFielders()
	updateOutRunners()
end

local OUT_BUBBLE_SIZE = 6

function drawScoreboard()
	gfx.setDrawOffset(0, 0)
	local y = Screen.H * 0.95
	local x = Screen.W * 0.05

	gfx.setLineWidth(1)
	gfx.setColor(gfx.kColorBlack)
	gfx.fillRect(x - 15, y - 44, 73, 55)

	gfx.setColor(gfx.kColorWhite)
	for i = outs, 2 do
		gfx.drawCircleAtPoint(x + (i * 2.5 * OUT_BUBBLE_SIZE), y, OUT_BUBBLE_SIZE)
	end
	for i = 0, (outs - 1) do
		gfx.fillCircleAtPoint(x + (i * 2.5 * OUT_BUBBLE_SIZE), y, OUT_BUBBLE_SIZE)
	end

	local originalDrawMode = gfx.getImageDrawMode()
	gfx.setImageDrawMode(playdate.graphics.kDrawModeInverted)
	gfx.drawText("Home: " .. homeScore, x - 7, y - 40, 100, 20)
	gfx.drawText("Away: " .. awayScore, x - 7, y - 25, 100, 20)
	gfx.setImageDrawMode(originalDrawMode)
end

function playdate.update()
	updateGameState()
	playdate.graphics.animation.blinker.updateAll()

	gfx.clear()

	if ball.x < BallOffscreen then
		-- TODO: Show baserunning minimap when panning?
		local offsetX, offsetY = getDrawOffset(Screen.W, Screen.H, ball.x, ball.y)
		gfx.setDrawOffset(offsetX, offsetY)
	end

	GrassBackground:draw(-400, -240)

	gfx.setColor(gfx.kColorBlack)
	gfx.setLineWidth(2)

	gfx.drawCircleAtPoint(ball.x, ball.y, ball.size)

	local fielderDanceHeight = FielderDanceAnimator:currentValue()
	for _, fielder in pairs(fielders) do
		gfx.fillRect(fielder.x, fielder.y - fielderDanceHeight, 14, 25)
	end

	if currentMode == Modes.batting then
		gfx.setLineWidth(5)
		gfx.drawLine(batBase.x, batBase.y, batTip.x, batTip.y)
	end

	if playdate.isCrankDocked() then -- or (crankChange < 2 and currentMode == Modes.running) then
		playdate.ui.crankIndicator:draw()
	end

	-- TODO? Change blip speed depending on runner speed?
	for _, runner in pairs(runners) do
		-- TODO? Scale sprites down as y increases
		PlayerImageBlipper:draw(false, runner.x, runner.y)
	end
	for _, runner in pairs(outRunners) do
		PlayerFrown:draw(runner.x, runner.y)
	end

	drawScoreboard()
end

init()
